#include <iostream>
#include <thread>
#include <atomic>
#include <chrono>
#include <vector>
#include <unordered_map>
#include <string.h>
#include <assert.h>
#include <tt_cplus_sdk.h>
#include "enums/OrderSide.h"
#include "enums/TimeInForce.h"
#include "enums/OrderType.h"
#include "enums/ExecType.h"
#include "enums/ProductType.h"
#include <iomanip>
#include <sstream>
#include <condition_variable>

#define ACCOUNT_ID 98829 //000000

std::mutex mutex;
std::condition_variable sdkReadyCondition;

void OutputAlgoDefintion(ttsdk::AlgoDefinitionPtr& algo)
{
    std::cout << "Downloaded algo definition for " << algo->GetName() << " version:" << algo->GetAlgoVersion() << std::endl;  
    std::cout << " Export Values:" << std::endl;
    for (uint32_t i = 0; i < algo->GetExportValueCount(); ++i)
    {
        auto e = algo->GetExportValueDefinition(i);
        std::cout << "  - displayName:  " << e->GetDisplayName() << " id: " << e->GetId() << " type: " << ttsdk::ToString(e->GetDataType()) << " blockName:" << e->GetBlockName() << " Connector:" << e->GetConnector() << std::endl;
    }
    std::cout << " Parameters:" << std::endl;
    for (uint32_t i = 0; i < algo->GetParameterCount(); ++i)
    {
        auto e = algo->GetParameterDefinition(i);
        std::cout << " - displayName: " << e->GetDisplayName() << " id: " << e->GetName() << " type  " << ttsdk::ToString(e->GetDataType()) << std::endl;
    }
}

class PlaceAlgoStrategy : public ttsdk::IOrderEventHandler, public ttsdk::IPriceEventHandler
{
public:
    // params that will be set for an iceberg test 
    const uint16_t childTifIndex = 0;
    const uint16_t parentTifIndex = 1;
    const uint16_t discValIndex = 2;
    const uint16_t discValTypeIndex = 3;

    PlaceAlgoStrategy(const char* alias, ttsdk::InstrumentPtr instrument)
        : alias_(alias), instrument_(instrument), running_(false) {};
    virtual ~PlaceAlgoStrategy()
    {
        Stop();
    };

    void Stop()
    {
        std::cout << "Stopping strategy for " << alias_ << std::endl;
        if (running_) 
        {
            if (order_)
            {
                // cancel if working
                auto execRpt = order_->GetCurrentState();
                if (execRpt && execRpt->IsWorking())
                {
                    if (exportsSubscribed_)
                    {
                        exportsSubscribed_ = false;
                        order_->UnSubscribeExportValues();
                    }
                    SendCancel();
                }
                else
                {
                    order_->Unsubscribe();
                }
            }
            if (subId_ != 0)
            {
                ttsdk::UnsubscribePrices(subId_);
            }
            else
            {
                running_ = false;
            }            
        } 
    };

    void Start()
    {
        std::cout << "Starting strategy for " << alias_ << std::endl;
        if (!instrument_)
        {
            std::cout << "Unable to start algo - instrument not specified." << std::endl;
            return;
        }

        //Get the Algo definition
        ttsdk::AlgoDefinition::ResponseCode algoRespCode;
        algo_ = ttsdk::GetAlgoDefinition(alias_.c_str(), algoRespCode);
        if (!algo_)
        {
            std::cout << "Unable to start algo - unable to load algo " <<  alias_ << std::endl;
            return;
        }
        //output
        OutputAlgoDefintion(algo_);
        // create the export mapping
        uint32_t exCount = algo_->GetExportValueCount();
        for (uint32_t i = 0; i < exCount; ++i)
        {
            auto e = algo_->GetExportValueDefinition(i);
            exportMapping_.insert(std::make_pair(std::string(e->GetId()), std::string(e->GetDisplayName())));
        }
        
        running_ = true;
        if (strcmp(alias_.c_str(), "TimeAndSales") == 0)
        {
            PlaceOrder();
        }
        else
        {
            subId_ = ttsdk::SubscribePrices(instrument_, 10, true, true, false, this);
        }
    }

    virtual void OnUnsubscribed(const uint64_t subId)
    {
        running_ = false;
        std::cout << "OnUnsubscribed from prices. subId=" << subId << std::endl;
    };
    virtual void OnDisconnect(const uint64_t subId, const ttsdk::InstrumentPtr&, const ttsdk::PriceEventType)
    {
        std::cout << "OnDisconnected from prices. subId=" << subId << std::endl;
    };
    virtual void OnPriceUpdate(const uint64_t, const ttsdk::InstrumentPtr&, const ttsdk::PriceSnap& snap)
    {
        if (!order_)
        {
            if (snap.bids.count > 0)
                best_ = snap.bids.levels[0].price;
            else if (!std::isnan(snap.last_trade_price))
                best_ = snap.last_trade_price;
            else if (!std::isnan(snap.settlement_price))
                best_ = snap.settlement_price;
            else
                return;
            PlaceOrder();
        }
    };
    virtual void OnError(const uint64_t subId, const ttsdk::InstrumentPtr& instrument,
        const ttsdk::SubscriptionError code, const char* error)
    {
        std::cout << "Error occurred subscribing for prices for instr=" << instrument->GetName()
            << " code=" << (int)code << " error=" << error << " subId=" << subId << std::endl;
    };
    inline void PlaceOrder()
    {
        //Create the Algo order
        order_ = ttsdk::CreateAlgoOrder(algo_, instrument_->GetInstrumentId());
        if (!order_)
        {
            std::cout << "Order creation failed. Stopping strategy." << std::endl;
            Stop();
            return;
        }

         order_->Subscribe(*this);

        profile_.request_id++;
        profile_.account_id = ACCOUNT_ID;
        profile_.price = instrument_->TickPriceDown(best_, 2);
        profile_.quantity = 20;
        profile_.type = ttsdk::OrderType::Limit;
        profile_.tif = ttsdk::TimeInForce::Day;
        profile_.side = ttsdk::OrderSide::Buy;

        //Set the user disconnect action (optional)
        profile_.user_disconnect_action = ttsdk::UserDisconnectAction::Leave;

        //Set the co location (optional)
        profile_.colocation = ttsdk::MarketId::CME;

        if (strcmp(alias_.c_str(), "TimeAndSales") == 0)
        {
           //For the export value algo used in this sample, no parameter need to be set
           if (algo_->GetExportValueCount() > 0)
            {
                //If the order is an Algo order, subscribe to the export values 
                if (order_->IsAlgoOrder() && !exportsSubscribed_)
                {
                    order_->SubscribeExportValues();
                    exportsSubscribed_ = true;
                }
                order_->SendNew(profile_);
                
                //To Pause an Algo
                PauseAlgo();
            }
        }
        else if (strcmp(alias_.c_str(), "TT_Iceberg") == 0)
        {
           //For the TT_Iceberg ALGO, below are the parameters that need to be set to set the disclosed qty and the disclosed qty type. 
           // setting 4 user parameters for this test
            profile_.params.resize(4);      
            for (uint32_t i = 0; i < algo_->GetParameterCount(); ++i)
            {
                //Get the parameter definition 
                auto e = algo_->GetParameterDefinition(i);

                const char* paramName = e->GetName();
                // The parameters starting with __ should be set directly on the profile. 
                if (strncmp(paramName, "__", strlen("__")) == 0)
                    continue;                
                    
                const char* paramDisplayName = e->GetDisplayName();
                // set the params we need to use
                if (strcmp(paramDisplayName, "ChildTIF") == 0)
                {
                    ttsdk::UserParameter& param = profile_.params[childTifIndex];
                    strcpy(param.name, paramName);
                    param.type = e->GetDataType();
                    param.v_int = 1;
                    std::cout << " Setting " << paramDisplayName << " to : " << param.v_int << std::endl;
                }
                else if (strcmp(paramDisplayName, "ParentTIF") == 0)
                {
                    ttsdk::UserParameter& param = profile_.params[parentTifIndex];
                    strcpy(param.name, paramName);
                    param.type = e->GetDataType();
                    param.v_int = 1;
                    std::cout << " Setting t" << paramDisplayName << " to : " << param.v_int << std::endl;
                }
                else if (strcmp(paramDisplayName, "DiscVal") == 0)
                {
                    //Display Qty value                    
                    ttsdk::UserParameter& param = profile_.params[discValIndex];
                    strcpy(param.name, paramName);
                    param.type = e->GetDataType();
                    param.v_double = 5.0;
                    std::cout << " Setting " << paramDisplayName << " to : " << param.v_double << std::endl;
                }
                else if (strcmp(paramDisplayName, "DiscValType") == 0)
                {
                    //Display qty type, its an enum : Qty = 1, Percent=2
                    ttsdk::UserParameter& param = profile_.params[discValTypeIndex];
                    strcpy(param.name, paramName);
                    param.type = e->GetDataType();                    
                    param.v_int = 1;
                    std::cout << " Setting " << paramDisplayName << " to : " << param.v_int << std::endl;
                }
            }
            order_->SendNew(profile_);
        }
        else
        {
            std::cout << "Unsupported algo. Stopping strategy." << std::endl;
            Stop();
        } 
    }

    void SendCancel()
    {
        profile_.request_id++;
        order_->SendCancel(profile_);
    }

    void PauseAlgo()
    {
        profile_.instruction = ttsdk::OrderProfile::Instruction::Hold;
        order_->SendCancel(profile_);
    }

    void ResumeAlgo()
    {
        profile_.instruction = ttsdk::OrderProfile::Instruction::Work;
        order_->SendCancel(profile_);
    }

    void SendQuantityChange()
    {
        // change the quantities
        std::cout << " Changing the quantity to 50 disclose 10" << std::endl;
        profile_.quantity = 50;
        profile_.params[discValIndex].v_double = 10.0;
        profile_.request_id++;
        order_->SendChange(profile_);
    }

    virtual void OnUnsubscribed(const char*) override
    {
        // order is now unsubscribed, unsub from the price stream
        if(subId_ > 0)
            ttsdk::UnsubscribePrices(subId_);
        else
            running_ = false;
            
        order_.reset();
        instrument_.reset();
    }

    virtual void OnExecutionReport(ttsdk::OrderPtr order, ttsdk::ExecutionReportPtr execRpt) override
    {
        switch (execRpt->GetExecType())
        {
        case ttsdk::ExecType::New:
            std::cout << "Order successfully entered in market. RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            if (strcmp(alias_.c_str(), "TT_Iceberg") == 0)
            {
                // change the qty of the iceberg 
                SendQuantityChange();
            }
            break;

        case ttsdk::ExecType::Restated:
            std::cout << "Order Restated  RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            break;
        case ttsdk::ExecType::Canceled:
            std::cout << "Order successfully canceled. RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << " [external=" << execRpt->IsExternalAction() <<
                "]" << std::endl;
            order_->Unsubscribe(); 
            break;
        case ttsdk::ExecType::Suspended:
            std::cout << "Order Suspended  RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;

            //Resume Algo
            ResumeAlgo();
            break;
        case ttsdk::ExecType::Replaced:
            std::cout << "Order replaced. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) <<
                " [external=" << execRpt->IsExternalAction() << "]" << std::endl;
            break;
        case ttsdk::ExecType::Trade:
            std::cout << "Order filled. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            // clear the order variable so a new order will be placed
            order_.reset();
            break;
        case ttsdk::ExecType::Rejected:
            std::cout << "Order filled. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            break;
        default:
            std::cout << "Unhandled event. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) <<
                " ExecutionType=" << ttsdk::ToString(execRpt->GetExecType()) << std::endl;
        }
    }
    virtual void OnReject(ttsdk::OrderPtr order, ttsdk::RejectResponsePtr rejResp) override
    {
        std::cout << "OnReject event. RequestId=" << rejResp->GetRequestId() <<
            " TTOrderId=" << order->GetOrderId() << " RejectReason=" << ttsdk::ToString(rejResp->GetRejectReason()) <<
            " Text=" << rejResp->GetText() << std::endl;
        order_.reset();
    }
    virtual void OnSendFailed(ttsdk::OrderPtr order, const ttsdk::OrderProfile& profile, const ttsdk::IOrderEventHandler::SendCode code) override
    {
        std::cout << "Order send failed. RequestId=" << profile.request_id <<
            " TTOrderId=" << order->GetOrderId() << " SendCode=" << (int)code << std::endl;
    }

    virtual void OnExportValuesUpdate(ttsdk::OrderPtr order, const ttsdk::ExportValues& exports) override
    {
        std::string order_id = order->GetOrderId();
        std::cout << "OnExportValuesUpdate for OrderId=" << order_id << std::endl;
        for (int j = 0; j < exports.count; ++j)
        {
            // need to do the mapping since we are showing the display name
            // could skip this if the raw namee (id) is known
            auto exportValue = exports.exportValues[j];
            assert(!exportValue.id.empty());
            auto iter = exportMapping_.find(exportValue.id);
            if (iter != exportMapping_.end())
            {
                std::cout << "  Name = " << iter->second;
                if (exportValue.v_double != NAN)
                    std::cout << ", value = " << exportValue.v_double << std::endl;
                else if (exportValue.v_int != -2147483648LL)
                    std::cout << ", value = " << exportValue.v_int << std::endl;
                else if (exportValue.v_timestamp != 0)
                        std::cout << ", value = " << exportValue.v_timestamp << std::endl;
                else if (exportValue.v_bool != false)
                    std::cout << ", value = " << exportValue.v_bool << std::endl;
                else if (exportValue.v_string != nullptr)
                    std::cout << ", value = " << exportValue.v_string << std::endl;
            }
        }
    }

    ttsdk::OrderProfile profile_;
    const std::string alias_;
    ttsdk::AlgoDefinitionPtr algo_;
    std::unordered_map<std::string, std::string> exportMapping_;
    ttsdk::InstrumentPtr instrument_;
    ttsdk::OrderPtr order_;
    std::atomic<bool> running_;
    uint64_t subId_ = 0;
    double best_ = 0.0;
    std::atomic<bool> exportsSubscribed_;
};

class MyOrderBookHandler : public ttsdk::IOrderBookEventHandler
{
public:
    MyOrderBookHandler() : IOrderBookEventHandler() {};
    virtual ~MyOrderBookHandler() {};
    virtual void OnExecutionReport(ttsdk::OrderPtr order, ttsdk::ExecutionReportPtr execRpt) override
    {
        std::cout << "MyOrderBookHandler::OnExecutionReport TTOrderId=" << order->GetOrderId() <<
            " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
    }
    virtual void OnEpiqUpdate(ttsdk::OrderPtr order, double epiq) override
    {
    }
    virtual void OnReject(ttsdk::OrderPtr order, ttsdk::RejectResponsePtr rejResp) override
    {
    }
    virtual void OnSendFailed(ttsdk::OrderPtr order, const ttsdk::OrderProfile& profile, const SendCode code) override
    {
    }
    virtual void OnUnsubscribed(const char* orderId) override
    {
    }
    virtual void OnAccountDownloadEnd(const uint64_t accountId) override
    {
    };
    virtual void OnOrderBookDownloadEnd() override
    {
        std::cout << "OnOrderBookDownloadEnd -- orders are ready to use" << std::endl;
        sdkReadyCondition.notify_one();
    };
    virtual void OnRiskReserved(const uint64_t instrumentId, const uint64_t accountId, const ttsdk::RiskSide side, const bool successful)
    {
    }
    virtual void OnRiskReleased(const uint64_t instrumentId, const uint64_t accountId, const ttsdk::RiskSide side, const bool successful)
    {
    }
};

class SDKEventHandler : public ttsdk::IEventHandler
{
public:
    SDKEventHandler() : IEventHandler() {};
    virtual ~SDKEventHandler() {};
    virtual void OnStatus(const ttsdk::IEventHandler::Status status)
    {
        std::cout << "SDKEventHandler::OnStatus (" << (uint32_t)status << ")" << std::endl;
    };
};

void DisplayHelp()
{
    // Use the following commands to control the application
    std::cout << "Commands: [q] Quit"
        << "\n          [h]  Display help"
        << "\n          [d]  Get Algo Definition"
        << "\n          [i]  Start TT Iceberg Algo Order"
        << "\n          [x]  Start TimeAndSales Algo"
        << std::endl;
}

int main(int argc, char* argv[])
{
    // this example is loading the app key secret from the command line or environment rather than hard coding.
    ttsdk::Environment env = ttsdk::Environment::UatCert;
    std::string sEnv("UatCert");
    std::string app_key;
    for (int i = 1; i < argc; ++i)
    {
        std::string arg = argv[i];
        if ((arg == "-k") || (arg == "--key"))
        {
            if (i + 1 < argc)
            {
                i++;
                app_key = argv[i];
            }
            else
            {
                std::cerr << "--key option requires one argument." << std::endl;
                return 1;
            }
        }
        else if ((arg == "-e") || (arg == "--env"))
        {
            if (i + 1 < argc)
            {
                i++;
                sEnv = argv[i];
                if (sEnv == "ProdLive") env = ttsdk::Environment::ProdLive;
                else if (sEnv == "ProdSim") env = ttsdk::Environment::ProdSim;
                else if (sEnv == "UatCert") env = ttsdk::Environment::UatCert;
                else if (sEnv == "DevCert") env = ttsdk::Environment::DevCert;
                else {
                    std::cerr << "Invalid env argument: " << sEnv << std::endl;
                    return 1;
                }
            }
            else
            {
                std::cerr << "--env option requires one argument." << std::endl;
                return 1;
            }
        }
    }
    if (app_key.empty())
    {
        auto env_app_key = std::getenv("TT_APP_KEY");
        if (!env_app_key || strcmp(env_app_key, "") == 0)
        {
            std::cout << "app_key is invalid. " << std::endl;
            return 1;
        }
        app_key = env_app_key;
    }
    std::cout << "using app_key: " << app_key << " in " << sEnv << std::endl;

    // these are the options for the SDK
    ttsdk::TTSDKOptions options;
    // set the environment the app needs to run (where your machine is located)
    options.environment = env;
    // set your app secret Key here. It looks like: 00000000-0000-0000-0000-000000000000:00000000-0000-0000-0000-000000000000
    options.app_key_secret = app_key.c_str();

    //Enable Synthetics to send synthetic orders
    options.enable_synthetics = true;

    MyOrderBookHandler myOBHandler;
    SDKEventHandler sdkObserver;
    if (!ttsdk::Initialize(options, &sdkObserver, &myOBHandler))
    {
        std::cout << "Unable to initialize SDK!" << std::endl;
        return -1;
    }
    // wait for the SDK to be initialized as ready to handle price requests
    std::unique_lock<std::mutex> lock(mutex);
    if (sdkReadyCondition.wait_for(lock, std::chrono::seconds(300)) == std::cv_status::timeout)
    {
        std::cout << "Timeout waiting for SDK to initialize!" << std::endl;
        return -1;
    }

    //Find the Instrument 
    ttsdk::Instrument::ResponseCode respCode;
    ttsdk::MarketId market = ttsdk::MarketId::CME;
    const char* product = "GE";
    ttsdk::ProductType productType = ttsdk::ProductType::Future;
    const char* alias = "GE Mar29";

    ttsdk::InstrumentPtr instrument = ttsdk::GetInstrument(market, product, productType, alias, respCode);

    if (!instrument)
    {
        std::cout << "Unable to get instrument market: " << ttsdk::ToString(market) << " product : " << product << " type : " << ttsdk::ToString(productType) << " alias : " << alias << " Stopping strategy." << std::endl;
        return -1;
    }

    std::vector< std::shared_ptr<PlaceAlgoStrategy>> myAlgoStrategies;

    std::string command;
    
    DisplayHelp();

    while (std::cin >> command)
    {
        if (command == "h")
        {
            DisplayHelp();
        }

        else if (command == "q")
        {
            std::cout << "Quitting...\n";
            break;
        }

        else if (command == "d")
        {
            std::string tmp;
            std::getline(std::cin, tmp);
            std::string algoName;

            std::cout << "Enter Algo Name:";
            std::getline(std::cin, algoName);
            std::cout << "Loading algo " << algoName << std::endl;

            ttsdk::AlgoDefinition::ResponseCode algoRespCode;
            ttsdk::AlgoDefinitionPtr theAlgo_ = ttsdk::GetAlgoDefinition(algoName.c_str(), algoRespCode);
            if (!theAlgo_)
            {
                std::cout << "Unable to load algo " << algoName;
            }
            else
            {
                std::cout << "Found algo " << algoName;
                OutputAlgoDefintion(theAlgo_);
            }
        }

        else if (command == "i")
        {
            auto strat = std::make_shared<PlaceAlgoStrategy>("TT_Iceberg", instrument);
            strat->Start();
            myAlgoStrategies.push_back(std::move(strat));
        }
        else if (command == "x")
        {
            auto strat = std::make_shared<PlaceAlgoStrategy>("TimeAndSales", instrument);
            strat->Start();
            myAlgoStrategies.push_back(std::move(strat));
        }
        

        // yield control back to OS to avoid hard spinning
        std::this_thread::yield();
    }

    bool running(true);
    while (running)
    {
        running = false;

        for (auto s : myAlgoStrategies)
        {
            if (s->running_)
            {
                s->Stop();
                running = true;
            }
        }
        if (running)
        {
            std::cout << "Waiting for running strategy " << std::endl;
            std::this_thread::sleep_for(std::chrono::milliseconds(150));
        }
    }

    myAlgoStrategies.clear();

    ttsdk::Shutdown();
}